class Collection(object):

    @classmethod
    def fetch_instruments_by_tag(cls, client, tag):
        base_url = "https://api.robinhood.com/midlands/tags/tag/"
        url = "{}{}/".format(base_url, tag.lower())
        res = client.get(url)
        result = (res["instruments"] if "instruments" in res else [])
        return result



from fast_arrow.exceptions import TradeExecutionError
from fast_arrow.util import is_max_date_gt

import uuid
import json


class OptionOrder(object):

    @classmethod
    def all(cls, client, **kwargs):
        """
        fetch all option positions
        """
        max_date = kwargs['max_date'] if 'max_date' in kwargs else None
        max_fetches = \
            kwargs['max_fetches'] if 'max_fetches' in kwargs else None

        url = 'https://api.robinhood.com/options/orders/'
        data = client.get(url)
        results = data["results"]

        if is_max_date_gt(max_date, results[-1]['updated_at'][0:10]):
            return results
        if max_fetches == 1:
            return results

        fetches = 1
        while data["next"]:
            fetches = fetches + 1
            data = client.get(data["next"])
            results.extend(data["results"])
            if is_max_date_gt(max_date, results[-1]['updated_at'][0:10]):
                return results
            if max_fetches and (fetches >= max_fetches):
                return results
        return results

    @classmethod
    def humanize_numbers(cls, option_orders):
        results = []
        for oo in option_orders:
            keys_to_humanize = ["processed_premium"]
            coef = (1.0 if oo["direction"] == "credit" else -1.0)
            for k in keys_to_humanize:
                if oo[k] is None:
                    continue
                oo[k] = float(oo[k]) * coef
            results.append(oo)
        return results

    @classmethod
    def unroll_option_legs(cls, client, option_orders):
        '''
        unroll option orders like this,
        https://github.com/joshfraser/robinhood-to-csv/blob/master/csv-options-export.py
        '''

        #
        # @TODO write this with python threats to make concurrent HTTP requests
        #

        results = []

        for oo in option_orders:
            for index, leg in enumerate(oo['legs']):
                for execution in leg['executions']:
                    order = dict()

                    keys_in_question = ['legs', 'price', 'type', 'premium',
                                        'processed_premium',
                                        'response_category', 'cancel_url']
                    for k, v in oo.items():
                        if k not in keys_in_question:
                            order[k] = oo[k]

                    order['order_type'] = oo['type']

                    contract = client.get(leg['option'])
                    order['leg'] = index+1
                    order['symbol'] = contract['chain_symbol']
                    order['strike_price'] = contract['strike_price']
                    order['expiration_date'] = contract['expiration_date']
                    order['contract_type'] = contract['type']

                    for k, v in leg.items():
                        if k not in ['id', 'executions']:
                            order[k] = leg[k]

                    coef = (-1.0 if leg['side'] == 'buy' else 1.0)
                    order['price'] = float(execution['price']) * 100.0 * coef
                    order['execution_id'] = execution['id']

                    results.append(order)
        return results

    @classmethod
    def submit(cls, client, direction, legs, price, quantity, time_in_force,
               trigger, order_type, run_validations=True):
        '''
        params:
        - client
        - direction
        - legs
        - price
        - quantity
        - time_in_force
        - trigger
        - order_type
        - run_validations. default = True
        '''

        if run_validations:
            assert(direction in ["debit", "credit"])
            assert(type(price) is str)
            assert(type(quantity) is int)
            assert(time_in_force in ["gfd", "gtc"])
            assert(trigger in ["immediate"])
            assert(order_type in ["limit", "market"])
            assert(cls._validate_legs(legs) is True)

        account_url = ("https://api.robinhood.com/accounts/{}/".
            format(client.account_id))

        payload = json.dumps({
            "account": account_url,
            "direction": direction,
            "legs": legs,
            "price": price,
            "quantity": quantity,
            "time_in_force": time_in_force,
            "trigger": trigger,
            "type": order_type,
            "override_day_trade_checks": False,
            "override_dtbp_checks": False,
            "ref_id": str(uuid.uuid4())
        })

        request_url = "https://api.robinhood.com/options/orders/"
        data = client.post(request_url, payload=payload)
        return data

    @classmethod
    def _validate_legs(cls, legs):
        for leg in legs:
            assert("option" in leg)
            assert(leg["position_effect"] in ["open", "close"])

            # @TODO research required formatting
            # assert(leg["ratio_quantity"])

            assert(leg["side"] in ["buy", "sell"])
        return True

    @classmethod
    def get(cls, client, option_order_id):
        request_url = \
            "https://api.robinhood.com/options/orders/{}/".format(
                option_order_id)
        return client.get(request_url)

    @classmethod
    def cancel(cls, client, cancel_url):
        result = client.post(cancel_url)
        if result == {}:
            return True
        else:
            return False

    @classmethod
    def replace(cls, client, option_order, new_price):
        result = cls.cancel(client, option_order["cancel_url"])
        if not result:
            msg = "OptionOrder.replace() did not cancel previous OptionOrder"
            raise TradeExecutionError(msg)

        account_url = ("https://api.robinhood.com/accounts/{}/".
            format(client.account_id))

        payload = json.dumps({
            "account": account_url,
            "direction": option_order["direction"],
            "legs": option_order["legs"],
            "price": new_price,
            "quantity": option_order["quantity"],
            "time_in_force": option_order["time_in_force"],
            "trigger": option_order["trigger"],
            "type": option_order["type"],
            "override_day_trade_checks": False,
            "override_dtbp_checks": False,
            "ref_id": str(uuid.uuid4())
        })

        request_url = "https://api.robinhood.com/options/orders/"
        data = client.post(request_url, payload=payload)
        return data



# library util and client stuff
from fast_arrow.client import Client

from fast_arrow.exceptions import (
    AuthenticationError,
    NotImplementedError)

# options
from fast_arrow.resources.option_chain import OptionChain
from fast_arrow.resources.option_event import OptionEvent
from fast_arrow.resources.option_order import OptionOrder
from fast_arrow.resources.option_marketdata import OptionMarketdata
from fast_arrow.resources.option_position import OptionPosition
from fast_arrow.resources.option import Option

# option_strategies
from fast_arrow.option_strategies.vertical import Vertical as Vertical
from fast_arrow.option_strategies.iron_condor import IronCondor as IronCondor

# stocks
from fast_arrow.resources.stock_order import StockOrder
from fast_arrow.resources.stock_marketdata import StockMarketdata
from fast_arrow.resources.stock_position import StockPosition
from fast_arrow.resources.stock import Stock

# user
from fast_arrow.resources.user import User
from fast_arrow.resources.account import Account
from fast_arrow.resources.portfolio import Portfolio

# search
from fast_arrow.resources.collection import Collection

# dividend
from fast_arrow.resources.dividend import Dividend

import warnings
warnings.simplefilter('always', DeprecationWarning)



from fast_arrow.util import chunked_list


class OptionMarketdata(object):

    @classmethod
    def quote_by_instrument_id(cls, client, _id):
        return cls.quotes_by_instrument_ids(client, [_id])[0]

    @classmethod
    def quotes_by_instrument_ids(cls, client, ids):
        base_url = "https://api.robinhood.com/options/instruments/"
        id_urls = ["{}{}/".format(base_url, _id) for _id in ids]
        return cls.quotes_by_instrument_urls(client, id_urls)

    @classmethod
    def quote_by_instrument_url(cls, client, url):
        return cls.quotes_by_instrument_urls(client, [url])[0]

    @classmethod
    def quotes_by_instrument_urls(cls, client, urls):
        results = []
        for _urls in chunked_list(urls, 50):
            url = "https://api.robinhood.com/marketdata/options/"
            params = {"instruments": ",".join(_urls)}
            data = client.get(url, params=params)
            if data and "results" in data:
                partial_results = data["results"]
                while ("next" in data and data["next"]):
                    data = client.get(data["next"])
                    partial_results.extend(data["results"])
                results.extend(partial_results)
        return results

    @classmethod
    def historical_quote_by_id(cls, client, _id, span="year"):
        return cls.historical_quotes_by_ids(client, [_id], span)[0]

    @classmethod
    def historical_quotes_by_ids(cls, client, ids, span="year"):
        base_url = "https://api.robinhood.com/options/instruments/"
        urls = ["{}{}/".format(base_url, _id) for _id in ids]
        return cls.historical_quotes_by_urls(client, urls)

    @classmethod
    def historical_quote_by_url(cls, client, url, span="year"):
        return cls.historical_quotes_by_urls(client, [url], span)[0]

    @classmethod
    def historical_quotes_by_urls(cls, client, urls, span="year"):
        possible_intervals = {
            "day": "5minute",
            "week": "10minute",
            "year": "day",
            "5year": "week"}
        assert span in possible_intervals.keys()
        interval = possible_intervals[span]
        results = []
        request_url = \
            "https://api.robinhood.com/marketdata/options/historicals/"
        for _urls in chunked_list(urls, 5):
            params = {"span": span, "interval": interval,
                      "instruments": ",".join(_urls)}
            data = client.get(request_url, params=params)
            if data and data["results"]:
                results.extend(data["results"])
        return results



from fast_arrow.resources.stock_marketdata import StockMarketdata


class Stock(object):

    @classmethod
    def fetch(cls, client, symbol):
        """
        fetch data for stock
        """
        assert(type(symbol) is str)

        url = ("https://api.robinhood.com/instruments/?symbol={0}".
               format(symbol))
        data = client.get(url)
        return data["results"][0]

    @classmethod
    def mergein_marketdata_list(cls, client, stocks):
        ids = [x["id"] for x in stocks]
        mds = StockMarketdata.quote_by_instruments(client, ids)
        mds = [x for x in mds if x]

        results = []
        for s in stocks:
            # @TODO optimize this so it's better than O(n^2)
            md = [md for md in mds if md['instrument'] == s['url']]
            if len(md) > 0:
                md = md[0]
                md_kv = {
                    "ask_price": md["ask_price"],
                    "bid_price": md["bid_price"],
                }
                merged_dict = dict(list(s.items()) + list(md_kv.items()))
            else:
                merged_dict = dict(list(s.items()))
            results.append(merged_dict)
        return results

    @classmethod
    def all(cls, client, symbols):
        """"
        fetch data for multiple stocks
        """
        params = {"symbol": ",".join(symbols)}
        request_url = "https://api.robinhood.com/instruments/"

        data = client.get(request_url, params=params)
        results = data["results"]

        while data["next"]:
            data = client.get(data["next"])
            results.extend(data["results"])
        return results



import pandas as pd


class Vertical(object):

    @classmethod
    def gen_df(cls, options, width, spread_type="call", spread_kind="buy"):
        """
        Generate Pandas Dataframe of Vertical

        :param options: python dict of options.
        :param width: offset for spread. Must be integer.
        :param spread_type: call or put. defaults to "call".
        :param spread_kind: buy or sell. defaults to "buy".
        """
        assert type(width) is int
        assert spread_type in ["call", "put"]
        assert spread_kind in ["buy", "sell"]

        # get CALLs or PUTs
        options = list(filter(lambda x: x["type"] == spread_type, options))

        coef = (1 if spread_type == "put" else -1)
        shift = width * coef

        df = pd.DataFrame.from_dict(options)
        df['expiration_date'] = pd.to_datetime(
            df['expiration_date'], format="%Y-%m-%d")
        df['adjusted_mark_price'] = pd.to_numeric(df['adjusted_mark_price'])
        df['strike_price'] = pd.to_numeric(df['strike_price'])

        df.sort_values(["expiration_date", "strike_price"], inplace=True)

        for k, v in df.groupby("expiration_date"):
            sdf = v.shift(shift)

            df.loc[v.index, "strike_price_shifted"] = sdf["strike_price"]
            df.loc[v.index, "delta_shifted"] = sdf["delta"]
            df.loc[v.index, "volume_shifted"] = sdf["volume"]
            df.loc[v.index, "open_interest_shifted"] = sdf["open_interest"]
            df.loc[v.index, "instrument_shifted"] = sdf["instrument"]
            df.loc[v.index, "adjusted_mark_price_shift"] = \
                sdf["adjusted_mark_price"]

            if spread_kind == "sell":
                df.loc[v.index, "margin"] = \
                    abs(sdf["strike_price"] - v["strike_price"])
            else:
                df.loc[v.index, "margin"] = 0.0

            if spread_kind == "buy":
                df.loc[v.index, "premium_adjusted_mark_price"] = (
                    v["adjusted_mark_price"] - sdf["adjusted_mark_price"])
            elif spread_kind == "sell":
                df.loc[v.index, "premium_adjusted_mark_price"] = (
                    sdf["adjusted_mark_price"] - v["adjusted_mark_price"])
        return df



from fast_arrow.util import chunked_list
from fast_arrow.resources.option_marketdata import OptionMarketdata
from fast_arrow.exceptions import NotImplementedError


class Option(object):

    @classmethod
    def fetch_by_url(cls, client, url):
        return cls.instrument_by_urls(client, [url])

    @classmethod
    def fetch_by_urls(cls, client, urls):
        raise NotImplementedError("Option.instrument_by_url")

    @classmethod
    def fetch_by_ids(cls, client, ids):
        results = []
        params = {"ids": ",".join(ids)}
        request_url = "https://api.robinhood.com/options/instruments/"
        data = client.get(request_url, params=params)
        results = data["results"]
        while data["next"]:
            data = client.get(data["next"])
            results.extend(data["results"])
        return results

    @classmethod
    def fetch_by_id(cls, bearer, _id):
        return cls.fetch_by_ids(bearer, [_id])

    # deprecate me
    @classmethod
    def fetch(cls, client, _id):
        """
        fetch by instrument id
        """
        return cls.fetch_list(client, [_id])[0]

    #
    # @TODO depricate me
    #
    @classmethod
    def fetch_list(cls, client, ids):
        """
        fetch instruments by ids
        """
        results = []
        request_url = "https://api.robinhood.com/options/instruments/"

        for _ids in chunked_list(ids, 50):

            params = {"ids": ",".join(_ids)}
            data = client.get(request_url, params=params)
            partial_results = data["results"]

            while data["next"]:
                data = client.get(data["next"])
                partial_results.extend(data["results"])
            results.extend(partial_results)

        return results

    @classmethod
    def in_chain(cls, client, chain_id, expiration_dates=[]):
        """
        fetch all option instruments in an options chain
        - expiration_dates = optionally scope
        """

        request_url = "https://api.robinhood.com/options/instruments/"
        params = {
            "chain_id": chain_id,
            "expiration_dates": ",".join(expiration_dates)
        }

        data = client.get(request_url, params=params)
        results = data['results']

        while data['next']:
            data = client.get(data['next'])
            results.extend(data['results'])
        return results

    @classmethod
    def mergein_marketdata_list(cls, client, options):
        ids = [x["id"] for x in options]
        mds = OptionMarketdata.quotes_by_instrument_ids(client, ids)
        mds = [x for x in mds if x]

        results = []
        for o in options:
            # @TODO optimize this so it's better than O(n^2)
            md = [md for md in mds if md['instrument'] == o['url']]
            if len(md) > 0:
                md = md[0]
                # there is no overlap in keys, so it's fine to do a merge
                merged_dict = dict(list(o.items()) + list(md.items()))
            else:
                merged_dict = dict(list(o.items()))
            results.append(merged_dict)

        return results



import numpy as np
from decimal import Decimal


class IronCondor(object):

    @classmethod
    def sort_by_strike_price(cls, options):
        return sorted(options, key=(lambda x: x['strike_price']))

    @classmethod
    def gen_leg(cls,
                option_url, side, position_effect="open", ratio_quantity=1):
        assert side in ["buy", "sell"]
        assert position_effect in ["open", "close"]
        return {
            "side": side,
            "option": option_url,
            "position_effect": position_effect,
            "ratio_quantity": ratio_quantity
        }

    @classmethod
    def max_bid_ask_spread(cls, ic_options):
        max_spread = Decimal(0.0)
        for x in ic_options:
            ask = Decimal(x["ask_price"])
            bid = Decimal(x["bid_price"])
            spread = ask - bid
            if spread > max_spread:
                max_spread = spread
        return float(max_spread)

    @classmethod
    def strings_to_np_array(cls, strings):
        x = np.array(strings)
        deltas = x.astype(np.float)
        return np.nan_to_num(deltas)

    @classmethod
    def generate_by_deltas(cls, options,
                           width, put_inner_lte_delta, call_inner_lte_delta):
        """
        totally just playing around ideas for the API.

        this IC sells
        - credit put spread
        - credit call spread

        the approach
        - set width for the wing spread (eg, 1, ie, 1 unit width spread)
        - set delta for inner leg of the put credit spread (eg, -0.2)
        - set delta for inner leg of the call credit spread (eg, 0.1)
        """

        raise Exception("Not Implemented starting at the 0.3.0 release")

        #
        # put credit spread
        #
        put_options_unsorted = list(
            filter(lambda x: x['type'] == 'put', options))
        put_options = cls.sort_by_strike_price(put_options_unsorted)

        deltas_as_strings = [x['delta'] for x in put_options]
        deltas = cls.strings_to_np_array(deltas_as_strings)

        put_inner_index = np.argmin(deltas >= put_inner_lte_delta) - 1
        put_outer_index = put_inner_index - width

        put_inner_leg = cls.gen_leg(
            put_options[put_inner_index]["instrument"], "sell")
        put_outer_leg = cls.gen_leg(
            put_options[put_outer_index]["instrument"], "buy")

        #
        # call credit spread
        #
        call_options_unsorted = list(
            filter(lambda x: x['type'] == 'call', options))
        call_options = cls.sort_by_strike_price(call_options_unsorted)

        deltas_as_strings = [x['delta'] for x in call_options]
        x = np.array(deltas_as_strings)
        deltas = x.astype(np.float)
        # because deep ITM call options have a delta that comes up as NaN,
        # but are approximately 0.99 or 1.0, I'm replacing Nan with 1.0
        # so np.argmax is able to walk up the index until it finds
        # "call_inner_lte_delta"
        # @TODO change this so (put credit / call credit) spreads work the same
        where_are_NaNs = np.isnan(deltas)
        deltas[where_are_NaNs] = 1.0

        call_inner_index = np.argmax(deltas <= call_inner_lte_delta)
        call_outer_index = call_inner_index + width

        call_inner_leg = cls.gen_leg(
            call_options[call_inner_index]["instrument"], "sell")
        call_outer_leg = cls.gen_leg(
            call_options[call_outer_index]["instrument"], "buy")

        legs = [put_outer_leg, put_inner_leg, call_inner_leg, call_outer_leg]

        #
        # price calcs
        #
        price = (
            - Decimal(put_options[put_outer_index]['adjusted_mark_price'])
            + Decimal(put_options[put_inner_index]['adjusted_mark_price'])
            + Decimal(call_options[call_inner_index]['adjusted_mark_price'])
            - Decimal(call_options[call_outer_index]['adjusted_mark_price'])
        )

        #
        # provide max bid ask spread diff
        #
        ic_options = [
            put_options[put_outer_index],
            put_options[put_inner_index],
            call_options[call_inner_index],
            call_options[call_outer_index]
        ]

        max_bid_ask_spread = cls.max_bid_ask_spread(ic_options)

        return {"legs": legs, "price": price,
                "max_bid_ask_spread": max_bid_ask_spread}



from fast_arrow import util
from fast_arrow.resources.option import Option
from fast_arrow.resources.option_marketdata import OptionMarketdata
from fast_arrow.util import is_max_date_gt


class OptionPosition(object):

    @classmethod
    def all(cls, client, **kwargs):
        """
        fetch all option positions
        """
        max_date = kwargs['max_date'] if 'max_date' in kwargs else None
        max_fetches = \
            kwargs['max_fetches'] if 'max_fetches' in kwargs else None

        url = 'https://api.robinhood.com/options/positions/'
        params = {}
        data = client.get(url, params=params)
        results = data["results"]

        if is_max_date_gt(max_date, results[-1]['updated_at'][0:10]):
            return results
        if max_fetches == 1:
            return results

        fetches = 1
        while data["next"]:
            fetches = fetches + 1
            data = client.get(data["next"])
            results.extend(data["results"])
            if is_max_date_gt(max_date, results[-1]['updated_at'][0:10]):
                return results
            if max_fetches and (fetches >= max_fetches):
                return results
        return results

    @classmethod
    def append_marketdata(cls, client, option_position):
        """
        Fetch and merge in Marketdata for option position
        """
        return cls.append_marketdata_list(client, [option_position])[0]

    @classmethod
    def mergein_marketdata_list(cls, client, option_positions):
        """
        Fetch and merge in Marketdata for each option position
        """
        ids = cls._extract_ids(option_positions)
        mds = OptionMarketdata.quotes_by_instrument_ids(client, ids)

        results = []
        for op in option_positions:
            # @TODO optimize this so it's better than O(n^2)
            md = [x for x in mds if x['instrument'] == op['option']][0]
            # there is no overlap in keys so this is fine
            merged_dict = dict(list(op.items()) + list(md.items()))
            results.append(merged_dict)
        return results

    @classmethod
    def mergein_instrumentdata_list(cls, client, option_positions):
        ids = cls._extract_ids(option_positions)
        idatas = Option.fetch_list(client, ids)

        results = []
        for op in option_positions:
            idata = [x for x in idatas if x['url'] == op['instrument']][0]
            # there is an overlap in keys,
            # {'chain_symbol', 'url', 'type', 'created_at', 'id',
            #    'updated_at', 'chain_id'}
            # @TODO this is ugly. let's fix it later
            # alternative method,
            #   wanted_keys = ['strike_price']
            #   idata_subset = \
            #       dict((k, idata[k]) for k in wanted_keys if k in idata)
            merge_me = {
                "option_type": idata["type"],
                "strike_price": idata["strike_price"],
                "expiration_date": idata["expiration_date"],
                "min_ticks": idata["min_ticks"]
            }
            merged_dict = dict(list(op.items()) + list(merge_me.items()))
            results.append(merged_dict)

        return results

    @classmethod
    def humanize_numbers(cls, option_positions):
        results = []
        for op in option_positions:
            keys_to_humanize = [
                "quantity",
                "delta",
                "theta",
                "gamma",
                "vega",
                "rho"]

            coef = (1.0 if op["type"] == "long" else -1.0)

            for k in keys_to_humanize:
                if op[k] is None:
                    continue
                op[k] = float(op[k]) * coef

            if op["type"] == "long":
                op["chance_of_profit"] = op["chance_of_profit_long"]
            else:
                op["chance_of_profit"] = op["chance_of_profit_short"]

            results.append(op)

        return results

    @classmethod
    def _extract_ids(cls, option_positions):
        ids = []
        for op in option_positions:
            _id = util.get_last_path(op["option"])
            ids.append(_id)
        return ids



from fast_arrow import util
from fast_arrow.resources.option import Option


class OptionEvent(object):

    @classmethod
    def all(cls, client):
        """
        fetch all option positions
        """
        url = 'https://api.robinhood.com/options/events/'
        params = {}
        data = client.get(url, params=params)
        results = data["results"]
        while data["next"]:
            data = client.get(data["next"])
            results.extend(data["results"])
        return results

    @classmethod
    def mergein_instrumentdata_list(cls, client, option_events):
        results = []
        ids = [util.get_last_path(oe['option']) for oe in option_events]
        idatas = Option.fetch_by_ids(client, ids)
        for oe in option_events:
            idata = [x for x in idatas if x['url'] == oe['option']][0]
            merge_me = {
                "option_type": idata["type"],
                "symbol": idata["chain_symbol"]
            }
            merged_dict = dict(list(oe.items()) + list(merge_me.items()))
            results.append(merged_dict)
        return results

    @classmethod
    def humanize_numbers(cls, option_events):
        results = []
        for oe in option_events:
            keys_to_humanize = ["total_cash_amount"]

            coef = 0.0
            if oe["type"] == "exercise":
                coef = 1.0
            elif oe["type"] == "assignment":
                coef = -1.0

            for k in keys_to_humanize:
                if oe[k] is None:
                    continue
                oe[k] = float(oe[k]) * (coef)
            results.append(oe)

        return results



from fast_arrow.util import chunked_list
import deprecation
from fast_arrow.version import VERSION


class StockMarketdata(object):

    @classmethod
    def quote_by_instrument(cls, client, _id):
        return cls.quote_by_instruments(client, [_id])[0]

    @classmethod
    def quote_by_instruments(cls, client, ids):
        """
        create instrument urls, fetch, return results
        """
        base_url = "https://api.robinhood.com/instruments"
        id_urls = ["{}/{}/".format(base_url, _id) for _id in ids]
        return cls.quotes_by_instrument_urls(client, id_urls)

    @classmethod
    def quotes_by_instrument_urls(cls, client, urls):
        """
        fetch and return results
        """
        instruments = ",".join(urls)
        params = {"instruments": instruments}
        url = "https://api.robinhood.com/marketdata/quotes/"
        data = client.get(url, params=params)
        results = data["results"]
        while "next" in data and data["next"]:
            data = client.get(data["next"])
            results.extend(data["results"])
        return results

    @classmethod
    @deprecation.deprecated(
        deprecated_in="0.3.1", removed_in="0.4", current_version=VERSION,
        details="Use 'historical_quote_by_symbol' instead")
    def historical(cls, client, symbol, span="year", bounds="regular"):
        return cls.historical_quote_by_symbol(client, symbol, span, bounds)

    @classmethod
    def historical_quote_by_symbol(cls, client, symbol, span="year",
                                   bounds="regular"):
        datas = cls.historical_quote_by_symbols(client, [symbol], span, bounds)
        return datas[0]

    @classmethod
    def historical_quote_by_symbols(cls, client, symbols, span="year",
                                    bounds="regular"):
        possible_intervals = {
            "day": "5minute",
            "week": "10minute",
            "year": "day",
            "5year": "week"}
        assert span in possible_intervals.keys()
        interval = possible_intervals[span]
        assert bounds in ["regular"]

        request_url = "https://api.robinhood.com/marketdata/historicals/"

        results = []
        for _symbols in chunked_list(symbols, 25):
            params = {"span": span, "interval": interval, "bounds": bounds,
                      "symbols": ",".join(_symbols)}
            data = client.get(request_url, params=params)
            if data and data["results"]:
                results.extend(data["results"])
        return results



class OptionChain(object):

    @classmethod
    def fetch(cls, client, _id, symbol):
        """
        fetch option chain for instrument
        """
        url = "https://api.robinhood.com/options/chains/"
        params = {
            "equity_instrument_ids": _id,
            "state": "active",
            "tradability": "tradable"
        }
        data = client.get(url, params=params)

        def filter_func(x):
            return x["symbol"] == symbol
        results = list(filter(filter_func, data["results"]))
        return results[0]



import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass into pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = '-n auto'

    def run_tests(self):
        import shlex
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


version_contents = {}
with open("fast_arrow/version.py", "r", encoding='utf-8') as f:
    exec(f.read(), version_contents)


with open("README.md", "r") as f:
    long_description = f.read()


deps = [
    'datetime',
    'deprecation',
    'pathlib2',
    'requests>=2.20.0',
    'pandas>=0.23.2',
    'numpy',
    'yarl',
    'urllib3>=1.24.2'
]


test_deps = [
    'pipenv',
    'pytest',
    'pytest-cov',
    'detox',
    'flake8',
    'vcrpy'
]


setup(name='fast_arrow',
    version=version_contents['VERSION'],
    description='A simple yet robust API client for Robinhood',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Weston Platter',
    author_email='westonplatter@gmail.com',
    url='https://github.com/westonplatter/fast_arrow/',
    license='MIT License',
    python_requires=">=3.5",
    packages=[
        'fast_arrow',
        'fast_arrow.resources',
        'fast_arrow.option_strategies'
    ],
    package_data={'fast_arrow': ['ssl_certs/certs.pem']},
    install_requires=deps,
    tests_require=test_deps,
    cmdclass={'test': PyTest},
    project_urls={
        'Issue Tracker': 'https://github.com/westonplatter/fast_arrow/issues',
        'Source Code': 'https://github.com/westonplatter/fast_arrow',
    }
)



class Portfolio(object):

    @classmethod
    def historical(cls, client, account="", span="year", bounds="regular"):
        possible_intervals = {
            "day": "5minute",
            "week": "10minute",
            "year": "day",
            "5year": "week"}
        assert span in possible_intervals.keys()
        interval = possible_intervals[span]
        assert bounds in ["regular", "trading"]

        base_request_url = "https://api.robinhood.com/portfolios/historicals"
        request_url = "{}/{}/".format(base_request_url, account)
        params = {"span": span, "interval": interval, "bounds": bounds}
        data = client.get(request_url, params=params)
        return data



import os
import requests
from fast_arrow.util import get_last_path
from fast_arrow.resources.account import Account
from fast_arrow.exceptions import AuthenticationError
from fast_arrow.exceptions import NotImplementedError
from fast_arrow.exceptions import AuthDataError

HTTP_ATTEMPTS_MAX = 2

CLIENT_ID = "c82SH0WZOsabOXGP2sxqcj34FxkvfnWRZBKlBjFS"

class Client(object):

    def __init__(self, auth_data):
        required_keys = [
            "account_id",
            "access_token",
            "refresh_token",
            "device_token"
        ]

        for k in required_keys:
            if k not in auth_data:
                msg = "Expected key={} not in auth_data".format(k)
                raise AuthDataError(msg)

        self.auth_data = auth_data
        self.account_id = auth_data["account_id"]
        self.access_token = auth_data["access_token"]
        self.refresh_token = auth_data["refresh_token"]
        self.device_token = auth_data["device_token"]
        certs_path = 'ssl_certs/certs.pem'
        self.certs = os.path.join(os.path.dirname(__file__), certs_path)


    def get(self, url=None, params=None, retry=True):
        '''
        Execute HTTP GET
        '''
        headers = self._gen_headers(self.access_token, url)
        attempts = 1
        while attempts <= HTTP_ATTEMPTS_MAX:
            try:
                res = requests.get(url,
                                   headers=headers,
                                   params=params,
                                   timeout=15,
                                   verify=self.certs)
                res.raise_for_status()
                return res.json()
            except requests.exceptions.RequestException as e:
                attempts += 1
                if res.status_code in [400, 401]:
                    raise e
                elif retry and res.status_code in [403]:
                    self.relogin_oauth2()

    def post(self, url=None, payload=None, retry=True):
        '''
        Execute HTTP POST
        '''
        headers = self._gen_headers(self.access_token, url)
        attempts = 1
        while attempts <= HTTP_ATTEMPTS_MAX:
            try:
                res = requests.post(url,
                                    headers=headers,
                                    data=payload,
                                    timeout=15,
                                    verify=self.certs)
                res.raise_for_status()
                if res.headers['Content-Length'] == '0':
                    return None
                else:
                    return res.json()
            except requests.exceptions.RequestException as e:
                attempts += 1
                if res.status_code in [400, 429]:
                    raise e
                elif retry and res.status_code in [403]:
                    self.relogin_oauth2()

    def _gen_headers(self, bearer, url):
        '''
        Generate headders, adding in Oauth2 bearer token if present
        '''
        headers = {
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate",
            "Accept-Language": ("en;q=1, fr;q=0.9, de;q=0.8, ja;q=0.7, " +
                                "nl;q=0.6, it;q=0.5"),
            "User-Agent": ("Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) " +
                           "AppleWebKit/537.36 (KHTML, like Gecko) " +
                           "Chrome/68.0.3440.106 Safari/537.36"),

        }
        if bearer:
            headers["Authorization"] = "Bearer {0}".format(bearer)
        if url == "https://api.robinhood.com/options/orders/":
            headers["Content-Type"] = "application/json; charset=utf-8"
        return headers

    def current_auth_data(self):
        auth_data = {
            "account_id": self.account_id,
            "access_token": self.access_token,
            "refresh_token": self.refresh_token,
            "device_token": self.device_token,
        }
        return auth_data

    def __set_account_info(self):
        account_urls = Account.all_urls(self)
        if len(account_urls) > 1:
            msg = ("fast_arrow 'currently' does not handle " +
                   "multiple account authentication.")
            raise NotImplementedError(msg)
        elif len(account_urls) == 0:
            msg = "fast_arrow expected at least 1 account."
            raise AuthenticationError(msg)
        else:
            self.account_url = account_urls[0]
            self.account_id = get_last_path(self.account_url)
            self.authenticated = True

    def relogin_oauth2(self):
        '''
        (Re)login using the Oauth2 refresh token
        '''
        url = "https://api.robinhood.com/oauth2/token/"
        data = {
            "grant_type": "refresh_token",
            "device_token": self.device_token,
            "refresh_token": self.refresh_token,
            "scope": "internal",
            "client_id": CLIENT_ID,
            "expires_in": 86400,
        }
        data = self.post(url, payload=data, retry=False)
        self.access_token = data["access_token"]
        self.refresh_token = data["refresh_token"]
        self.scope = data["scope"]
        return True

    def logout_oauth2(self):
        '''
        Logout for given Oauth2 bearer token
        '''
        url = "https://api.robinhood.com/oauth2/revoke_token/"
        data = {
            "client_id": CLIENT_ID,
            "token": self.refresh_token,
        }
        res = self.post(url, payload=data)
        if res is None or res == {}:
            self.account_id = None
            self.account_url = None
            self.access_token = None
            self.refresh_token = None
            self.mfa_code = None
            self.scope = None
            self.authenticated = False
            return True
        else:
            raise AuthenticationError("fast_arrow could not log out.")




class Dividend(object):

    @classmethod
    def all(cls, client):
        """"
        fetch data for dividends
        """
        url = "https://api.robinhood.com/dividends/"
        data = client.get(url)
        results = data["results"]
        while data["next"]:
            data = client.get(data["next"])
            results.extend(data["results"])
        return results
